#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "e_lib.h"

unsigned short *const output_flag = (void *)0x4000;
e_mutex_t *mutex_output = (void *)0x4008;

e_coreid_t id;
unsigned row, col;
e_memseg_t message_buf;
int retval = EXIT_SUCCESS;

int main(void) {

  id = e_get_coreid();
  e_coords_from_coreid(id, &row, &col);

  // Attach to the shm segment
  if (E_OK != e_shm_attach(&message_buf, "outbuf")) {
    retval = EXIT_FAILURE;
    goto done;
  }

  // init mutex on core (0,0)
  if (row == 0 && col == 0) {
    e_mutex_init(0, 0, mutex_output, MUTEXATTR_NULL);
  }

  const char Msg[] = "(%d,%d) Parallel Hello World!\n";
  char tmp[128];

  while (1) {
    // check message_flag
    if (*output_flag == 1) {
      // tell host that core should busy wait
      *output_flag = 2;
      // wait for host clearance
      while (*output_flag != 0) { /*busy wait*/
      }
    }

    // get mutual access to message buffer
    e_mutex_lock(0, 0, mutex_output);
    // create message
    snprintf(tmp, sizeof(tmp), Msg, row, col);
    // check if buffer has enough space left
    if (strnlen((char *)message_buf.ephy_base, message_buf.size) + strlen(tmp) <
        message_buf.size) {
      e_write(&message_buf, tmp, 0, 0, strlen((char *)message_buf.ephy_base),
              strlen(tmp) + 1);
    }
    e_mutex_unlock(0, 0, mutex_output);
  }

done:
  return retval;
}
