#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <e-hal.h>

#define _SIZE (512)

unsigned short zero = 0;
unsigned short one = 1;
unsigned short tmp = 0;
char outbuf[_SIZE];
char terminate = '\0';

int main(int argc, char *argv[]) {
#ifdef DEBUG
  fprintf(stdout, "Starting host application ...\n");
#endif
  int retval = EXIT_SUCCESS;
  unsigned i, j;
  e_platform_t platform;
  e_epiphany_t dev;
  e_mem_t message_buf;

  if (E_OK != e_init(NULL)) {
    fprintf(stderr, "Epiphany HAL initialization failed\n");
    return EXIT_FAILURE;
  }

  if (E_OK != e_reset_system()) {
    fprintf(stderr, "Epiphany system reset failed\n");
    retval = EXIT_FAILURE;
    goto err_out1;
  }

  if (E_OK != e_get_platform_info(&platform)) {
    fprintf(stderr, "Failed to get Epiphany platform info\n");
    retval = EXIT_FAILURE;
    goto err_out1;
  }

  // allocate buffer
  if (E_OK != e_shm_alloc(&message_buf, "outbuf", _SIZE)) {
    fprintf(stderr, "Failed to allocate shared memory. Error is %s\n",
            strerror(errno));
    retval = EXIT_FAILURE;
    goto err_out2;
  }

  // initialize core memory
  for (i = 0; i < platform.rows; i++) {
    for (j = 0; j < platform.cols; j++) {
      // init message_flag
      e_write(&dev, i, j, 0x4000, &zero, sizeof(zero));
    }
  }

  if (E_OK != e_open(&dev, 0, 0, platform.rows, platform.cols)) {
    fprintf(stderr, "Failed to open device.\n");
    retval = EXIT_FAILURE;
    goto err_out2;
  }

  if (E_OK != e_load_group("e_parallelOutbuf.elf", &dev, 0, 0, platform.rows,
                           platform.cols, E_FALSE)) {
    fprintf(stderr, "Failed to load e_shm_test.elf\n");
    retval = EXIT_FAILURE;
    goto err_out2;
  }

  usleep(10000);

  // Start all of the cores
  if (E_OK != e_start_group(&dev)) {
    fprintf(stderr, "Failed to start group.\n");
    retval = EXIT_FAILURE;
    goto err_out2;
  }

  // TODO add barrier after each core initialization
  usleep(10000);

  // start main loop n times
  unsigned n = 3;
  while (n > 0) {
    n--;
#ifdef DEBUG
    fprintf(stdout, "Host loop n=%d\n", n);
#endif
    // notify all cores to busy wait
    for (i = 0; i < platform.rows; i++) {
      for (j = 0; j < platform.cols; j++) {
        e_write(&dev, i, j, 0x4000, &one, sizeof(one));
      }
    }

    // wait till all cores responded
    for (i = 0; i < platform.rows; i++) {
      for (j = 0; j < platform.cols; j++) {
        while (tmp != 2) {
          e_read(&dev, i, j, 0x4000, &tmp, sizeof(tmp));
        }
        // reset tmp
        tmp = 0;
      }
    }

    // read message_buf
    e_read(&message_buf, 0, 0, 0x0, outbuf, _SIZE);
    fprintf(stdout, "%s\n",
            outbuf); // TODO remove outbuf and output with message_buf.phy_base

    // clear message_buf
    e_write(&message_buf, 0, 0, 0x0, &terminate, sizeof(terminate));

    // notify all cores to continue
    for (i = 0; i < platform.rows; i++) {
      for (j = 0; j < platform.cols; j++) {
        e_write(&dev, i, j, 0x4000, &zero, sizeof(zero));
      }
    }
  }

  e_close(&dev);

err_out2:
  e_shm_release("outbuf");
err_out1:
  e_finalize();

  return retval;
}
